@@ -2,10 +2,15 @@
 /* Statnet is protected under the GNU Public License (GPL2). */
 /* Author: Jeroen Baekelandt (jeroenb@igwe.vub.ac.be)       */
 
+#ifdef OLDLINUX
 #include <ncurses.h>
+#else
+#include <curses.h>
+#endif
 #include <string.h>
 #include <stdarg.h>
 #include "curs.h"
+#include "netwatch.h"
 extern int colour;
 
 void
@@ -40,13 +45,9 @@ clrscr ()
 {
   int i;
   clear ();
-  if (colour)
-    {
-/*      bkgdset(COLOR_PAIR(4));   */
-      attrset (COLOR_PAIR (4));
-      for (i = 0; i < LINES; i++)
-	mvchgat (i, 0, COLS, COLOR_PAIR (4), 0, NULL);
-    }
+  attrset (col4);
+  for (i = 0; i < LINES; i++)
+    mvchgat (i, 0, COLS, col4, 0, NULL);
   border (0, 0, 0, 0, 0, 0, 0, 0);
   refresh ();
 }
