@@ -1,11 +1,12 @@
 /*
- * Strobe v0.92 (c) 1994 *Proff* (proff@suburbia.apana.org.au), All
- * rights reserved.
+ * Strobe (c) 1995 Julian Assange (proff@suburbia.net),
+ * All rights reserved.
  *
- * use of this program is at your own, or others risk. the author is
- * not considered a member of either group. comprehend?
+ * $ cc strobe.c -o strobe
  */
 
+#define VERSION "1.03"
+
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -15,18 +16,21 @@
 #include <fcntl.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
+#ifdef _AIX
+#  include <sys/select.h>
+#endif
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netdb.h>
 #include <string.h>
 #include <errno.h>
 
-#if !defined(solaris) && !defined(linux) && !defined(__FreeBSD__) && !defined(__NetBSD__) && !defined(__GCC__)
+#if defined(solaris) || defined(linux) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(__GCC__)
+#  define fvoid void
+#else
 #  define fvoid
 extern int optind;
 extern char *optarg;
-#else
-#  define fvoid void
 #endif
 #define bool char
 
@@ -34,6 +38,8 @@ extern char *optarg;
 #  define INADDR_NONE ((unsigned long)-1)
 #endif
 
+#define port_t (unsigned short)
+
 /*
  * the below should be set via the Makefile, but if not...
  */
@@ -52,10 +58,14 @@ int a_timeout = 20;
 char *a_output = NULL;
 char *a_services = "strobe.services";
 char *a_input = NULL;
+/* char *a_prescan = NULL; */
 int a_start = 1;
 int a_end = 65535;
 int a_sock_max = 64;
 int a_abort = 0;
+int a_bindport = 0;
+char *a_bindaddr= NULL;
+struct in_addr bindaddr;
 bool f_linear = 0;
 bool f_verbose = 0;
 bool f_verbose_stats = 0;
@@ -63,27 +73,35 @@ bool f_fast = 0;
 bool f_stats = 0;
 bool f_quiet = 0;
 bool f_delete_dupes = 0;
+bool f_minimise = 0;
+bool f_dontgetpeername = 0;
 
-bool f_set;
-int connects;
-int hosts_done=-1;
-int attempts_done;
+int connects = 0;
+int hosts_done = 0;
+int attempts_done = 0;
+int attempts_outstanding = 0;
 struct timeval time_start;
 
+fd_set set_sel;
+fd_set set_sel_r;
+fd_set set_sel_w;
+
 int host_n;
 int Argc;
 char **Argv;
 
 FILE *fh_input;
 
-#define HO_ABORT 1
-#define HO_COMPLETING 2
+#define HO_ACTIVE 1
+#define HO_ABORT 2
+#define HO_COMPLETING 4
 
 struct hosts_s
 {
     char *name;
     struct in_addr in_addr;
-    unsigned int port;
+    int port;
+    int portlist_ent;
     struct timeval time_used;
     struct timeval time_start;
     int attempts;
@@ -93,16 +111,17 @@ struct hosts_s
     time_t notice_abort;
     int status;
 };
+struct hosts_s ho_initial;
 struct hosts_s *hosts;
 
 #define HT_SOCKET 1
 #define HT_CONNECTING 2
 
-struct htuple
+struct htuple_s
 {
     char *name;
     struct in_addr in_addr;
-    unsigned int port;
+    int port;
     int sfd;
     int status;
     struct timeval sock_start;
@@ -110,26 +129,49 @@ struct htuple
     struct hosts_s *host;
 };
 
-struct htuple ht_initial;
-struct htuple *attempt;
+struct htuple_s ht_initial;
+struct htuple_s *attempt;
 
 struct port_desc_s
 {
+    int port;
     char *name;
     char *portname;
     struct port_desc_s *next;
+    struct port_desc_s *next_port;
 };
 
-struct port_desc_s *port_descs[65536];
+struct port_desc_s **port_descs;
+
+int *portlist = NULL;
+int portlist_n = 0;
+
+char *
+Srealloc (ptr, len)
+  char *ptr;
+  int len;
+{
+    char *p;
+    int retries = 10;
+    while (!(p = ptr? realloc (ptr, len): malloc(len)))
+    {
+        if (!--retries)
+        {
+		perror("malloc");
+		exit(1);
+	}
+	if (!f_quiet)
+	   fprintf(stderr, "Smalloc: couldn't allocate %d bytes...sleeping\n", len);
+	sleep (2);
+    }
+    return p;
+}
 
 char *
 Smalloc (len)
   int len;
 {
-    char *p;
-    while (!(p = malloc (len)))
-	sleep (1);
-    return p;
+   return Srealloc (NULL, len);
 }
 
 fvoid
@@ -151,7 +193,7 @@ sock_unblock (sfd)
 }
 
 int
-timeval_subtract (result, x, y)
+timeval_subtract (result, x, y) /* from gnu c-lib info.texi */
   struct timeval *result, *x, *y;
 {
 /* Perform the carry for the later subtraction by updating y. */
@@ -177,7 +219,7 @@ return x->tv_sec < y->tv_sec;
 
 fvoid
 attempt_clear (h)
-  struct htuple *h;
+  struct htuple_s *h;
 {
     if (h->status & HT_SOCKET)
     {
@@ -195,8 +237,13 @@ attempt_clear (h)
 	if (h->port > h->host->attempts_highest_done)
 	    h->host->attempts_highest_done=h->port;
 	sock_unblock (h->sfd);
-	shutdown (h->sfd, 2);
+/*	shutdown (h->sfd, 2); */
 	close (h->sfd);
+        if (FD_ISSET(h->sfd, &set_sel))
+	{
+	     FD_CLR (h->sfd, &set_sel);
+	     attempts_outstanding--;
+	}
     }
     *h = ht_initial;
 }
@@ -217,18 +264,32 @@ attempt_init ()
 	attempt[n] = ht_initial;
 }
 
+fvoid
+hosts_init ()
+{
+    int n;
+    for (n = 0; n < a_sock_max; n++)
+	hosts[n] = ho_initial;
+}
+
+fvoid
+fdsets_init ()
+{
+    FD_ZERO(&set_sel_r); /* yes, we have to do this, despite the later */
+    FD_ZERO(&set_sel_w); /* assisgnments */
+    FD_ZERO(&set_sel);
+}
+
 int
 sc_connect (h)
-  struct htuple *h;
+  struct htuple_s *h;
 {
     struct sockaddr_in sa_in;
     int sopts1 = 1;
     struct linger slinger;
-    sa_in.sin_family = AF_INET;
-    sa_in.sin_addr = h->in_addr;
-    sa_in.sin_port = htons (h->port);
     if ((h->sfd = socket (PF_INET, SOCK_STREAM, 0)) == -1)
 	return 0;
+    memset(&sa_in, 0, sizeof(sa_in));
     h->status |= HT_SOCKET;
     gettimeofday(&(h->sock_start), NULL);
     sock_unblock (h->sfd);
@@ -236,6 +297,24 @@ sc_connect (h)
     setsockopt (h->sfd, SOL_SOCKET, SO_OOBINLINE, (char *) &sopts1, sizeof (sopts1));
     slinger.l_onoff = 0;	/* off */
     setsockopt (h->sfd, SOL_SOCKET, SO_LINGER, (char *) &slinger, sizeof (slinger));
+    sa_in.sin_family = AF_INET;
+    if (a_bindport)
+        sa_in.sin_port = a_bindport;
+    if (a_bindaddr)
+        sa_in.sin_addr = bindaddr;
+    if (a_bindaddr || a_bindport)
+        if (bind (h->sfd, (struct sockaddr *)&sa_in, sizeof(sa_in)) == -1)
+        {
+		fprintf(stderr, "couldn't bind %s : %d  ", a_bindaddr? a_bindaddr: "0.0.0.0", ntohs(a_bindport));
+		perror("");
+		if (errno == EACCES)
+			exit(1);
+		return 0;
+	}
+    setsockopt (h->sfd, SOL_SOCKET, SO_REUSEADDR, (char *) &sopts1, sizeof (sopts1));
+    sa_in.sin_addr = h->in_addr;
+    sa_in.sin_port = htons (h->port);
+
     if (connect (h->sfd, (struct sockaddr *) &sa_in, sizeof (sa_in)) == -1)
     {
 	switch (errno)
@@ -246,6 +325,7 @@ sc_connect (h)
 	case ETIMEDOUT:
 	case ECONNREFUSED:
 	case EADDRNOTAVAIL:
+	case EPERM:
 	    if (f_verbose)
 	    {
 		fprintf(stderr, "%s:%d ", h->name, h->port);
@@ -267,55 +347,62 @@ sc_connect (h)
     h->host->attempts++;
     h->status |= HT_CONNECTING;
     sock_block (h->sfd);
+    FD_SET(h->sfd, &set_sel);
+    attempts_outstanding++;
     return 1;
 }
 
 int
 gatherer_tcp (h)
-  struct htuple *h;
+  struct htuple_s *h;
 {
     struct port_desc_s *pd;
-    if ((pd = port_descs[h->port]))
-    {
-	    printf ("%-30s %-16s %5d/tcp %s\n", h->name, pd->portname, h->port, pd->name);
-	while (!f_delete_dupes && (pd=pd->next))
-	    printf ("#%-29s %-16s %5d/tcp %s\n", h->name, pd->portname, h->port, pd->name);
-    }
+    if (f_minimise)
+	printf ("%s\t%d\n", h->name, h->port);
     else
+    {
+    	if ((pd = port_descs[h->port]))
+        {
+    	    	printf ("%-30s %-16s %5d/tcp %s\n", h->name, pd->portname, h->port, pd->name);
+	    while (!f_delete_dupes && !f_minimise && (pd=pd->next))
+	        printf ("#%-29s %-16s %5d/tcp %s\n", h->name, pd->portname, h->port, pd->name);
+    	}
+    	else
 	    printf ("%-30s %-16s %5d/tcp unassigned\n", h->name, "unknown", h->port);
+    }
     h->host->connects++;
     connects++;
-        attempt_clear (h);
+    attempt_clear (h);
     return 1;
 }
 
 bool
 gather ()
 {
-    fd_set set_sel_r;
-    fd_set set_sel_w;
     struct timeval timeout;
-    struct htuple *h;
+    struct htuple_s *h;
     int n;
-    /* We regenerate the entire set rather than just use set=old_set
-     * because in some implimentations fs_set is just an array
-     */
-    FD_ZERO(&set_sel_r);
-    FD_ZERO(&set_sel_w);
-    for (f_set=n=0; n<a_sock_max; n++)
-    {
-        h=&attempt[n];
-	if (h->status & HT_CONNECTING)
-        {
-            FD_SET(h->sfd, &set_sel_r);
-            FD_SET(h->sfd, &set_sel_w);
-	    f_set=1;
-        }
-    }
-    if (!f_set) return 1;
+    int selected;
+    time_t tim;
+
+    if (!attempts_outstanding) return 1;
+    set_sel_r=set_sel_w=set_sel;
     timeout.tv_sec = 0;
-    timeout.tv_usec = 50;
-    switch (select (FD_SETSIZE, &set_sel_r, &set_sel_w, 0, &timeout))
+    timeout.tv_usec = 250000; /* 1/4 of a second */
+
+    selected = select (FD_SETSIZE, &set_sel_r, &set_sel_w, 0, &timeout);
+    /* Look for timeouts */
+
+    tim = time (NULL);
+    for ( n = 0 ; n < a_sock_max; n++ )
+    {
+      h = &attempt[n];
+      if ((h->status & HT_SOCKET) &&
+	  ((h->sock_start.tv_sec + h->timeout) < tim))
+ 	attempt_clear (h);
+    }
+
+    switch (selected)
     {
     case -1:
 	perror ("select");
@@ -323,7 +410,7 @@ gather ()
     case 0:
 	return 1;
     }
-    for (n = 0; n < a_sock_max; n++)
+    for (n = 0; selected && (n < a_sock_max); n++)
     {
 	h = &attempt[n];
 	if (h->status & HT_CONNECTING)
@@ -332,12 +419,18 @@ gather ()
 	    {
 		struct sockaddr_in in;
 		int len = sizeof (in);
+		selected--;
 		/* select() lies occasionaly
                  */
-		if (getpeername (h->sfd, (struct sockaddr *) &in, &len) == 0)
-		    gatherer_tcp (h);
+		if (!f_dontgetpeername) /* but solaris2.3 crashes occasionally ;-| */
+		{
+			if (getpeername (h->sfd, (struct sockaddr *) &in, &len) == 0)
+		    	    gatherer_tcp (h);
+			else
+		    	    attempt_clear (h);
+		}
 		else
-		    attempt_clear (h);
+		    	gatherer_tcp (h);
 	    }
 	}
     }
@@ -346,26 +439,20 @@ gather ()
 
 bool
 add_attempt (add)
-  struct htuple *add;
+  struct htuple_s *add;
 {
-    struct htuple *h;
+    struct htuple_s *h;
     static time_t oldtime;
-    int n;
+    static int n;
     for (;;)
     {
-	for (n = 0; n < a_sock_max; n++)
+	for (; n < a_sock_max; n++)
 	{
-	    time_t tim = time (0);
 	    h = &attempt[n];
 	    if (!h->status)
 		goto foundfree;
-	    if ((h->status & HT_SOCKET) &&
-		((h->sock_start.tv_sec + h->timeout) < tim))
-	    {
-		attempt_clear (h);
-		goto foundfree;
-	    }
 	}
+	n = 0;
 	gather ();
 	continue;
       foundfree:
@@ -375,9 +462,9 @@ add_attempt (add)
 	    gather ();
 	    continue;
 	}
-	if ((oldtime + 1) < time (0))
+	if ((oldtime + 1) < time (NULL))
 	{
-	    oldtime = time (0);
+	    oldtime = time (NULL);
 	    gather ();
 	}
 	break;
@@ -390,13 +477,13 @@ scatter (host, timeout)
   struct hosts_s *host;
   int timeout;
 {
-    static struct htuple add;
+    static struct htuple_s add;
     add = ht_initial;
     add.host = host;
     add.name = host->name;
     add.in_addr = host->in_addr;
     add.port = host->port;
-    add.timeout = a_timeout;
+    add.timeout = timeout;
     if (f_verbose)
 	fprintf (stderr, "attempting port=%d host=%s\n", add.port, add.name);
     add_attempt (&add);
@@ -408,11 +495,11 @@ wait_end (t)
   int t;
 {
     time_t st;
-    st = time (0);
-    while ((st + t) > time (0))
+    st = time (NULL);
+    while ((st + t) > time (NULL))
     {
 	gather ();
-	if (!f_set) break;
+	if (attempts_outstanding<1) break;
     }
 }
 
@@ -449,17 +536,19 @@ reread:
 	if (!fgets (lbuf, sizeof (lbuf), fh_input))
 	{
 	    fclose (fh_input);
-	    return NULL;
+            a_input = NULL;
+	    return next_host();
 	}
-	if (strchr("# \t\n", lbuf[0])) goto reread;
-	n = strcspn (lbuf, " \t\n");
+	if (strchr("# \t\n\r", lbuf[0])) goto reread;
+	n = strcspn (lbuf, " \t\n\r");
 	if (n)
 	    lbuf[n] = '\0';
 	return lbuf;
     }
-    if (++host_n >= Argc)
-	return NULL;
-    return Argv[host_n];
+    if ( host_n >= Argc )
+      return NULL;
+
+    return Argv[host_n++];
 }
 
 bool
@@ -469,7 +558,7 @@ host_init (h, name, nocheck)
   bool nocheck;
 {
     int n;
-    h->name = 0;
+    *h=ho_initial;
     h->in_addr = resolve (name);
     if (h->in_addr.s_addr == INADDR_NONE)
 	return 0;
@@ -487,14 +576,7 @@ host_init (h, name, nocheck)
     h->name = (char *) Smalloc (strlen (name) + 1);
     strcpy (h->name, name);
     h->port = a_start;
-    h->time_used.tv_sec
-      = h->time_used.tv_usec
-      = h->attempts_done 
-      = h->attempts
-      = h->connects
-      = h->notice_abort
-      = h->attempts_highest_done
-      = h->status = 0;
+    h->status = HO_ACTIVE;
     gettimeofday(&(h->time_start), NULL);
     return 1;
 }
@@ -506,8 +588,8 @@ host_clear (h)
     if (h->name)
     {
     	free (h->name);
-    	h->name = NULL;
     }
+    *h=ho_initial;
 }
 
 fvoid
@@ -559,22 +641,34 @@ bool skip_host(h)
     return 0;
 }
 
+int
+next_port (h)
+struct hosts_s *h;
+{
+    int n;
+    for (n = h->port; ++n <= a_end;)
+    {
+    	if (!f_fast) return n;
+	if (++h->portlist_ent>portlist_n) return -1;
+        return (portlist[h->portlist_ent-1]);
+    }
+    return -1;
+}
+
 fvoid
 scan_ports_linear ()
 {
     struct hosts_s host;
     char *name;
-    int n;
     while ((name = next_host ()))
     {
 	if (!host_init(&host, name, 1)) continue;
-	for (n = a_start; n <= a_end; n++)
+	for (;;)
 	{
-	    if (f_fast && !port_descs[n])
-		continue;
-	    host.port = n;
 	    scatter (&host, a_timeout);
 	    if (skip_host(&host)) break;
+	    if ((host.port = next_port(&host))==-1)
+		break;
 	}
 	wait_end (a_timeout);
 	if (f_verbose_stats)
@@ -587,12 +681,15 @@ scan_ports_linear ()
 /* Huristics:
  *  o  fast connections have priority == maximise bandwidth i.e 
  *     a port in the hand is worth two in the bush
+ *
  *  o  newer hosts have priority == lower ports checked more quickly
+ *
  *  o  all hosts eventually get equal "socket time" == despite
  *     priorities let no one host hog the sockets permanently
+ *
  *  o  when host usage times are equal (typically on or shortly after
  *     initial startup) distribute hosts<->sockets evenly rather than
- *     play a game of chaotic bifurcation ping-pong
+ *     play a game of chaotic bifurcatic ping-pong
  */
           
 fvoid
@@ -610,19 +707,29 @@ scan_ports_paralell ()
 	for (n = 0, smallest_cnt = 0xfffffff, smallest = NULL; n < a_sock_max; n++)
 	{
 	    h = &hosts[n];
+
+	    if (((h->status & HO_COMPLETING) &&
+                 (h->attempts_done == h->attempts)) ||
+                skip_host(h))
+	    {
+		if (f_verbose_stats) host_stats (h);
+		host_clear (h);
+	    }
+
 	    if (!h->name && ((name = next_host ())))
 		if (!host_init (h, name, 0))
 		{
 		    host_clear (h);
 		    continue;
 		}
+
 	    if (h->name)
 	    {
 		if (((h->time_used.tv_sec < smallest_val.tv_sec) ||
 		     ((h->time_used.tv_sec == smallest_val.tv_sec) &&
-		      (h->time_used.tv_usec <= h->time_used.tv_usec))) &&
-		    (((h->time_used.tv_sec!=smallest_val.tv_sec) &&
-		      (h->time_used.tv_usec!=smallest_val.tv_sec)) ||
+		      (h->time_used.tv_usec <= smallest_val.tv_usec))) &&
+		    (((h->time_used.tv_sec != smallest_val.tv_sec) &&
+		      (h->time_used.tv_usec != smallest_val.tv_usec)) ||
 		     (h->attempts < smallest_cnt)))
 	        {
 	  	    smallest_cnt = h->attempts;
@@ -631,27 +738,17 @@ scan_ports_paralell ()
 		 }
 	    }
 	}
+
 	if (smallest)
 	{
 	    if (!(smallest->status & HO_COMPLETING))
 	    {
 		scatter (smallest, a_timeout);
-	        if (f_fast)
-	  	    while ((++(smallest->port) < a_end) &&
-		       !port_descs[(smallest->port)]);
-	        else
-		    smallest->port++;
-	        if (smallest->port >= a_end) smallest->status|=HO_COMPLETING;
+		if ((smallest->port=next_port(smallest))==-1)
+	            smallest->status|=HO_COMPLETING;
 	    }
 	    else
 		gather();
-	    if (((smallest->status & HO_COMPLETING) &&
-                 (smallest->attempts_done == smallest->attempts)) ||
-                skip_host(smallest))
-	    {
-		if (f_verbose_stats) host_stats (smallest);
-		host_clear (smallest);
-	    }
 	}
     }
 }
@@ -663,16 +760,19 @@ loaddescs ()
     char lbuf[1024];
     char desc[256];
     char portname[17];
+    unsigned int port;
+    char *fn;
     char prot[4];
-    int port;
     prot[3]='\0';
-    if (!(fh = fopen (a_services, "r")) &&
-        !(fh = fopen (LIB_STROBE_SERVICES, "r")) &&
-        !(fh = fopen (ETC_SERVICES, "r")))
+    if (!(fh = fopen ((fn=a_services), "r")) &&
+        !(fh = fopen ((fn=LIB_STROBE_SERVICES), "r")) &&
+        !(fh = fopen ((fn=ETC_SERVICES), "r")))
     {
-	perror (a_services);
+	perror (fn);
 	exit (1);
     }
+    port_descs=(struct port_desc_s **) Smalloc(sizeof(struct port_descs_s *) * 65536);
+    memset(port_descs, 0, 65536);
     while (fgets (lbuf, sizeof (lbuf), fh))
     {
 	char *p;
@@ -683,57 +783,76 @@ loaddescs ()
 	desc[0]='\0';
 	if (sscanf (lbuf, "%16s %u %3s %255[^\r\n]", portname, &port, prot, desc) <3 || strcmp (prot, "tcp") || (port > 65535))
 	    continue;
-	pdp = (struct port_desc_s *) Smalloc (sizeof (*pd) + strlen (desc) + 1 + strlen (portname) + 1);
-	if ((pd = port_descs[port]))
+	pd = port_descs[port];
+	if (!pd)
 	{
-	    for (; pd->next; pd = pd->next);
-	    pd->next = pdp;
-	    pd = pd->next;
+	    portlist = (int *)Srealloc((char *)portlist, ++portlist_n*sizeof(int));
+	    portlist[portlist_n-1]=port;
 	}
-	else 
+	if (!f_minimise)
 	{
-	    pd = pdp;
+	    pdp = (struct port_desc_s *) Smalloc (sizeof (*pd) + strlen (desc) + 1 + strlen (portname) + 1);
+	    if (pd)
+	    {
+	        for (; pd->next; pd = pd->next);
+	        pd->next = pdp;
+	        pd = pd->next;
+	    } else 
+	    {
+	        pd = pdp;
+	        port_descs[port] = pd;
+	    } 
 	    pd->next = NULL;
-	    port_descs[port] = pd;
-	}
-	pd->name = (char *) (pd) + sizeof (*pd);
-	pd->portname = pd->name + strlen(desc)+1;
-	strcpy (pd->name, desc);
-	strcpy (pd->portname, portname);
+	    pd->name = (char *) (pd) + sizeof (*pd);
+	    pd->portname = pd->name + strlen(desc)+1;
+	    strcpy (pd->name, desc);
+	    strcpy (pd->portname, portname);
+	} else
+	    port_descs[port] = (struct port_desc_s *)-1;
     }
+    if (f_minimise) 
+        free (port_descs);
 }
 
 fvoid
 usage ()
 {
     fprintf (stderr, "\
-usage: strobe [-v(erbose)]\n\
-	      [-V(erbose_stats]\n\
-	      [-s(tatistics)]\n\
-	      [-q(uiet)]\n\
-              [-o output_file]\n\
-              [-b begin_port_n]\n\
-              [-e end_port_n]\n\
-              [-t timeout_n]\n\
-              [-n num_sockets_n]\n\
-              [-S services_file]\n\
-              [-i hosts_input_file]\n\
-              [-l(inear)]\n\
-              [-f(ast)]\n\
-              [-a abort_port_n]\n\
-              [host1 [...host_n]]\n");
+usage: %8s [options]\n\
+\t\t[-v(erbose)]\n\
+\t\t[-V(erbose_stats]\n\
+\t\t[-m(inimise)]\n\
+\t\t[-d(elete_dupes)]\n\
+\t\t[-g(etpeername_disable)]\n\
+\t\t[-s(tatistics)]\n\
+\t\t[-q(uiet)]\n\
+\t\t[-o output_file]\n\
+\t\t[-b begin_port_n]\n\
+\t\t[-e end_port_n]\n\
+\t\t[-p single_port_n]\n\
+\t\t[-P bind_port_n]\n\
+\t\t[-A bind_addr_n]\n\
+\t\t[-t timeout_n]\n\
+\t\t[-n num_sockets_n]\n\
+\t\t[-S services_file]\n\
+\t\t[-i hosts_input_file]\n\
+\t\t[-l(inear)]\n\
+\t\t[-f(ast)]\n\
+\t\t[-a abort_after_port_n]\n\
+\t\t[-M(ail_author)]\n\
+\t\t[host1 [...host_n]]\n", Argv[0]);
     exit (1);
 }
-
 int
 main (argc, argv)
   int argc;
   char **argv;
 {
-    char c;
-    if (argc < 2)
-	usage ();
-    while ((c = getopt (argc, argv, "o:dvVb:e:a:t:n:S:i:lfsq")) != -1)
+    int c;
+    Argc = argc;
+    Argv = argv;
+
+    while ((c = getopt (argc, argv, "o:dvVmgb:e:p:P:a:A:t:n:S:i:lfsqM")) != -1)
 	switch (c)
 	{
 	case 'o':
@@ -748,12 +867,33 @@ main (argc, argv)
 	case 'V':
 	    f_verbose_stats = 1;
 	    break;
+	case 'm':
+	    f_minimise = 1;
+	    break;
+	case 'g':
+	    f_dontgetpeername = 1;
+	    break;
 	case 'b':
 	    a_start = atoi (optarg);
 	    break;
 	case 'e':
 	    a_end = atoi (optarg);
 	    break;
+	case 'P':
+	    a_bindport = htons (atoi (optarg));
+	    break;
+	case 'A':
+	    a_bindaddr = optarg;
+	    bindaddr = resolve (a_bindaddr);
+	    if (bindaddr.s_addr == INADDR_NONE)
+	    {
+	    	perror(a_bindaddr);
+		exit(1);
+	    }
+            break;
+	case 'p':
+	    a_start = a_end = atoi (optarg);
+	    break;
 	case 'a':
 	    a_abort = atoi (optarg);
 	    break;
@@ -781,23 +921,48 @@ main (argc, argv)
         case 'q':
 	    f_quiet = 1;
 	    break;
+	case 'M':
+	    fprintf(stderr, "Enter mail to author below. End with ^D or .\n");
+	    system("mail strobe@suburbia.net");
+	    break;
 	case '?':
 	default:
-	    fprintf (stderr, "unknown option %s\n", argv[optind]);
+	    fprintf (stderr, "unknown option %s\n", argv[optind-1]);
 	    usage ();
 	    /* NOT_REACHED */
 	}
-    host_n = optind - 1;
+    host_n = optind;
+
     if (!f_quiet)
-        fprintf (stderr, "strobe (c) 1994 *Proff* All Rights Reserved.\n");
+        fprintf (stderr, "strobe %s (c) 1995 Julian Assange (proff@suburbia.net).\n", VERSION);
     if (a_input)
     {
-	if (!(fh_input = fopen (a_input, "r")))
-	{
-	    perror (a_input);
-	    exit (1);
+        if ( ! strcmp("-",a_input) ) { /* Use stdin as input file */
+	    fh_input = stdin;
+	  }
+	else {
+	  if (!(fh_input = fopen (a_input, "r")))
+	    {
+	      perror (a_input);
+	      exit (1);
+	    }
 	}
+    } else
+    {
+      switch ( argc - host_n ) { /* Number of hosts found on command line */
+      case 0:
+	fh_input = stdin;
+	a_input = "stdin"; /* Needed in "next_host()" */
+	break;
+      case 1:
+	f_linear = 1;
+	break;
+      }
     }
+
+    if ((fh_input==stdin) && !f_quiet)
+      fprintf (stderr, "Reading host names from stdin...\n");
+
     if (a_output)
     {
         int fd;
@@ -808,14 +973,16 @@ main (argc, argv)
 	}
 	dup2(fd, 1);
     }
-    Argc = argc;
-    Argv = argv;
-    if (!a_input && optind==argc-1)
-	f_linear=1;
-    attempt = (struct htuple *) Smalloc (a_sock_max * sizeof (struct htuple));
+    attempt = (struct htuple_s *) Smalloc (a_sock_max * sizeof (struct htuple_s));
+    attempt_init();
     if (!f_linear)
+    {
     	hosts = (struct hosts_s *) Smalloc (a_sock_max * sizeof (struct hosts_s));
-    loaddescs ();
+    	hosts_init();
+    }
+    if (!f_minimise || f_fast)
+    	loaddescs ();
+    fdsets_init();
     gettimeofday(&time_start, NULL);
     f_linear ? scan_ports_linear ():
  	       scan_ports_paralell ();
